from math import pow, sqrt
from typing import List


# naca constants
A0 =  0.2969
A1 = -0.1260
A2 = -0.3516
A3 =  0.2843


a0123 = lambda x: A0*sqrt(x) + A1*x + A2*pow(x,2) + A3*pow(x,3)


def a4(finite_te: bool) -> float:
    if finite_te:
        return -0.1015
    else:
        return -0.1036


def thickness_distribution(t: float, finite_te: bool, xc: list) -> List[float]:
    '''
    NACA-4 and 5 series airfoils share the same thickness distribution (y_t)
    '''
    _a4 = a4(finite_te)
    y_t = [5*t * (a0123(x) + _a4*pow(x,4)) for x in xc]

    # force the trailing edge to zero
    if not finite_te: y_t[-1] = 0
    return y_t


import unittest
import foillib as fl


class test_naca_4(unittest.TestCase):

    def setUp(self):
        self.x, self.y = fl.naca("0012")

    # ensure that the maxium thickness of the NACA0012 is in fact 12% of chord.
    def test_max_thickness(self):
        max_thickness = abs(min(self.y)) + max(self.y)
        self.assertAlmostEqual(max_thickness, 0.12, 4)

    # ensure that the default airfoil has a non-zero trailing edge thickness
    def test_trailing_edge_thickness(self):
        self.assertGreater(self.y[0], 0)
        self.assertLess(self.y[-1], 0)


class test_naca_5(unittest.TestCase):

    def setUp(self):
        self.x, self.y = fl.naca("24012")

    # ensure that the default airfoil has a non-zero trailing edge thickness
    def test_trailing_edge_thickness(self):
        self.assertGreater(self.y[0], 0)
        self.assertLess(self.y[-1], 0)


if __name__ == '__main__':
    unittest.main()


from math import cos, sin, sqrt
from typing import List, Tuple

from ._common import spacing
from ._types import SpacingMethod


def b2(t: float, p0: float, p1: float, p2: float) -> float:
    '''
    Quadratic definition
    
    t  : float
    p0 : float
    p1 : float
    p2 : float
    '''
    return p1 + (p0 - p1)*(1-t)**2 + (p2 - p1)*t**2


def b2_find_t(x: list, a: float, b: float, c:float) -> float:
    '''
    Find the roots: given x, solve for t
    '''
    d = abs(c*x + x*a - 2*x*b + b**2 - c*a)
    if a + c != 2 * b:
        t =  (a - b + sqrt(d)) / (c + a - 2*b)
    elif a == 2*b - c and b != c:
        t = -(-2*b + c + x) / (2*(b-c))
    return max(min(1, t), 0)


def b2_segment(x: list, p0: list, p1: list, p2:list) -> List[Tuple]:
    '''
    Single quadratic segment
    
    '''
    
    # determine which chord points fall within the current bezier segment
    x_seg = [_x for _x in x if p0[0] < _x <= p2[0]]
            
    # given the chord positions, determine the local segment positions (t)
    t = [b2_find_t(_x, p0[0], p1[0], p2[0]) for _x in x_seg]

    # compute the bezier y-values
    y = [b2(_t, p0[1], p1[1], p2[1]) for _t in t]
    
    # munge and return
    return list(zip(*[x_seg, y]))


def quadratic_bezier(x: List[float], cps: List[Tuple]) -> List[Tuple]:
    '''
    Piecewise Quadratic Bezier Curve
    
    x   : x-coordinate positions along chord (le -> te).
    cps : list of tuples representing the bezier controil points [(x1,y1), (x2,y2),...,(xn,yn)s]
    '''

    # compute the first segment
    curve = []
    x_mid = (cps[1][0] + cps[2][0]) / 2.0
    y_mid = (cps[1][1] + cps[2][1]) / 2.0
    curve += b2_segment(x, cps[0], cps[1], (x_mid, y_mid))
            
    # cycle through the intermediate segments
    for i in range(1, len(cps)-3):
        x_mid1 = (cps[i][0] + cps[i+1][0]) / 2.0
        y_mid1 = (cps[i][1] + cps[i+1][1]) / 2.0
        x_mid2 = (cps[i+1][0] + cps[i+2][0]) / 2.0
        y_mid2 = (cps[i+1][1] + cps[i+2][1]) / 2.0
        curve += b2_segment(x, (x_mid1, y_mid1), cps[i+1], (x_mid2, y_mid2))
        
    # compute final segment
    x_mid = (cps[-3][0] + cps[-2][0]) / 2.0
    y_mid = (cps[-3][1] + cps[-2][1]) / 2.0
    curve += b2_segment(x, (x_mid, y_mid), cps[-2], cps[-1])
        
    return curve


def bezier(cpts_le: List, cpts_top: List[Tuple], cpts_bot: List[Tuple]=None, te: float=0.00252, npts: int=100, spacing_method: SpacingMethod="cosine") -> List[Tuple]:
    '''
    The Bezier airfoil is comprised of two composite (piecewise) Bezier curves connected at the
    leading edge with G1 (tangent) continuity.
    
    cps_le  : 
    cps_top : 
    cps_bot : 
    te      : 
    npts    : 
    spacing_method : 
    '''
    
    # general input checks
    assert len(cpts_top) >= 2, "not enough control points specified for top surface..."
    assert len(cpts_le) in [1, 2], "leading edge radius has not been specified..."
    for v in cpts_le:
        assert v > 0, "leading edge radius must be definied with positive values"
    for i in range(len(cpts_top) - 1):
        assert cpts_top[i][0] < cpts_top[i+1][0], "top surface control point x-coordinates must be monotonically increasing"
    if cpts_bot != None:
        assert len(cpts_bot) >= 2, "not enough control points specified for bottom surface..."
        for i in range(len(cpts_bot) - 1):
            assert cpts_bot[i][0] < cpts_bot[i+1][0], "bottom surface control point x-coordinates must be monotonically increasing"
    
    # generate chord positions
    xc = spacing(npts, spacing_method)
    
    # munge leading edge control points
    if len(cpts_le) == 1:
        cpts_le = [abs(cpts_le[0])] * 2
    
    # top surface
    cpts_top = [(0, 0), (0,  cpts_le[0])] + cpts_top + [(1,  te / 2)]
    xy_top = quadratic_bezier(xc, cpts_top)
    
    # bottom surface
    if cpts_bot != None:
        cpts_bot = [(0, 0), (0,  -cpts_le[1])] + cpts_bot + [(1, -te / 2)]
        xy_bot = quadratic_bezier(xc, cpts_bot)
    else:
        xy_bot = [(x, -y) for x, y in xy_top]
    
    # assemble the entire curve
    xy = xy_top[::-1] + [(0,0)] + xy_bot
    
    return list(zip(*xy))


from math import atan, cos, pow, sin
from typing import List, Tuple

from ._common import spacing
from ._naca_common import thickness_distribution
from ._types import SpacingMethod


# tabluated camberline parameters
prk1 = {
    # non-reflexed (simple)
    0: {
        '210': {
            'p': 0.05, 'r': 0.0580, 'k1': 361.40
        },
        '220': {
            'p': 0.10, 'r': 0.1260, 'k1': 51.640
        },
        '230': {
            'p': 0.15, 'r': 0.2025, 'k1': 15.957
        },
        '240': {
            'p': 0.20, 'r': 0.2900, 'k1':  6.643
        },
        '250': {
            'p': 0.25, 'r': 0.3910, 'k1':  3.230
        }
    },
    
    # reflexed
    1: {
        '221': {
            'p': 0.05, 'r': 0.0580, 'k1': 361.40, 'k2/k1': 0.000764 
        },
        '231': {
            'p': 0.15, 'r': 0.2170, 'k1': 15.793, 'k2/k1': 0.00677 
        },
        '241': {
            'p': 0.20, 'r': 0.3180, 'k1':  6.520, 'k2/k1': 0.0303 
        },
        '251': {
            'p': 0.25, 'r': 0.4410, 'k1':  3.191, 'k2/k1': 0.1355 
        }
    }
}


def naca5(L: int, P: int, S: int, TT: int, npts: int, finite_te: bool, spacing_method=SpacingMethod) -> List[Tuple]:
    '''
    NACA-5 SERIES AIRFOIL -> "LPSXX"

    L  : Designed coefficient of lift multiplied by 3/20.
    P  : The position of maximum camber divided by 20.
    S  : 0 = normal camber line, 1 = reflexed camber line.
    XX : max thickness (t/c) in percent chord.
    '''
    
    # munge inputs
    t = float(TT) / 100.0
    profile = f'{L:d}{P:d}{S:d}'
    
    # generate chord positions
    xc = spacing(npts, spacing_method)
    
    # compute airfoil half-thickness
    yt = thickness_distribution(t, finite_te, xc)
    
    # determine interpolated camberline parameters
    if S == 0:
        p, r, k1= (
            prk1[S][profile]['p'],
            prk1[S][profile]['r'],
            prk1[S][profile]['k1']
        )
    else:
         p, r, k1, k2_k1 = (
            prk1[S][profile]['p'],
            prk1[S][profile]['r'],
            prk1[S][profile]['k1'],
            prk1[S][profile]['k2/k1']
        )
    
    # determine mean camber line (yc), camber gradient (dyc_dx), and theta
    xu, yu, xl, yl = [0] * len(xc), [0] * len(xc), [0] * len(xc), [0] * len(xc)
    for i, x in enumerate(xc):
        
        # non-reflex (standard)
        if S == 0:
            if x <= r:
                yc = (k1/6) * (pow(x,3) - 3*r*pow(x,2) + pow(r,2)*(3-r)*x)
                dyc_dx = (k1/6) * (3*pow(x,2) - 6*r*x + pow(r,2)*(3-r))
            else:
                yc = (k1*pow(r,3)/6) * (1 - x)
                dyc_dx = -(k1*pow(r,3)/6)
            
        # reflex
        if S == 1:
            if x <= r:
                yc = (k1/6) * (pow(x-r,3) - k2_k1*pow(1-r, 3)*x - pow(r,3)*x + pow(r,3))
                dyc_dx = (k1/6) * (3*pow(x-r,2) - k2_k1*pow(1-r,3) - pow(r,3))
            else:
                yc = (k1/6) * (k2_k1 * pow(x-r, 3) - k2_k1*pow(1-r, 3)*x - pow(r,3)*x + pow(r,3))
                dyc_dx = (k1/6) * (3 * k2_k1 * pow(x-r, 2) - k2_k1 * pow(1-r, 3) - pow(r,3))
            
        # compute surface coordiantes
        theta = atan(dyc_dx)
        xu[i] = x  - yt[i] * sin(theta)
        yu[i] = yc + yt[i] * cos(theta)
        xl[i] = x  + yt[i] * sin(theta)
        yl[i] = yc - yt[i] * cos(theta)
    
    x = xu[::-1] + [0] + xl
    y = yu[::-1] + [0] + yl
    return x, y



from math import cos, pi
from typing import List

from ._types import SpacingMethod


def spacing(npts: int, spacing_method: SpacingMethod = "cosine") -> List[float]:
    '''
    - Starts at the LE -> TE
    - The leading edge point (0,0) is not included.
    '''
    assert spacing_method in ['equal', 'cosine']
    if spacing_method == 'equal':
        return [x / npts for x in range(1, npts + 1)]
    elif spacing_method == 'cosine':
        return [(1 - cos(x / npts * pi)) / 2 for x in range(1, npts + 1)]


from math import atan, cos, pow, sin, sqrt
from typing import List, Tuple

from ._common import spacing
from ._naca_common import thickness_distribution
from ._types import SpacingMethod


def naca4(M: int, P: int, XX: int, npts: int, finite_te: bool, spacing_method: SpacingMethod) -> List[Tuple]:
    '''
    NACA-4 SERIES AIRFOIL -> "MPXX"
    
    M  : maximum value of the mean line in hundredths of chord.
    P  : chordwise position of max camber in tenths of chord.
    XX : max thickness (t/c) in percent chord.
    '''
    
    # munge thickness input
    t = float(XX) / 100.0
        
    # generate chord positions
    xc = spacing(npts, spacing_method)
    
    # compute airfoil half-thickness
    yt = thickness_distribution(t, finite_te, xc)
    
    # symmetrical airfoils
    if P == 0:
        x = xc[::-1] + [0] + xc
        y = yt[::-1] + [0] + [-v for v in yt]
        return x, y
    
    # cambered airfoils
    m = float(M) / 100.0
    p = float(P) /  10.0
    
    # determine mean camber line (yc), camber gradient (dyc_dx), and theta
    xu, yu, xl, yl = [0] * len(xc), [0] * len(xc), [0] * len(xc), [0] * len(xc)
    for i, x in enumerate(xc):
        if x <= p:
            yc = m / pow(p,2) * (2*p*x - pow(x,2))
            dyc_dx = 2*m / pow(p,2) * (p - x)
        else:
            yc = m / pow(1-p, 2) * ((1 - 2*p) + (2*p*x) - pow(x,2))
            dyc_dx = 2*m / pow(1-p,2) * (p - x)
            
        # compute surface coordiantes
        theta = atan(dyc_dx)
        xu[i] = x  - yt[i] * sin(theta)
        yu[i] = yc + yt[i] * cos(theta)
        xl[i] = x  + yt[i] * sin(theta)
        yl[i] = yc - yt[i] * cos(theta)
    
    # assemble final airfoil coordinates
    x = xu[::-1] + [0] + xl
    y = yu[::-1] + [0] + yl
    return x, y



from typing import List

from ._naca4 import naca4
from ._naca5 import naca5, prk1
from ._types import SpacingMethod


def naca(digits: str, npts=100, finite_te=True, spacing_method: SpacingMethod="cosine") -> List[tuple]:
    has_dash = '-' in digits
    assert len(digits) in [4, 5]
    
    # naca-4 series airfoils
    if len(digits)==4 and has_dash==False:
        return naca4(
            int(digits[0]), int(digits[1]), int(digits[2:]),
            npts, finite_te, spacing_method
        )
        
    # naca-5 series airfoils
    if len(digits)==5 and has_dash==False:
        assert digits[:3] in prk1[0].keys() or digits[:3] in prk1[1].keys(), "Unsupported design"
        assert digits[2] in ['0','1'], 'Ensure that the camber line is defined as: regular="1" or reflexed="2"'
        return naca5(
            int(digits[0]), int(digits[1]), int(digits[2]), int(digits[3:]),
            npts, finite_te, spacing_method
        )
    
    # naca-6 series airfoils
    # coming soon maybe...


